#!/usr/bin/env python
# encoding: utf-8
"""
NcML -> NetCDF code, heavily borrowed from davidh's polar2grid

Copyright © 2014-2017 University of Wisconsin Regents

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

from config import CSPP_GEO_GRBR_DEFAULT_COMPRESSION

import sys
import logging
import numpy
from xml.etree import cElementTree
from netCDF4 import Dataset

log = logging.getLogger('GRB-R')

# FIXME: a hack to work around some of the metadata having non-ASCII
# characters in it, is this really expected?


def unistr(s):
    try:
        return str(s)
    except:
        return s.encode('utf8')

_types = {
    "float": numpy.float32,
    "int": numpy.int32,
    "uint": numpy.uint32,
    None: unistr,
    "string": unistr,
    "short": numpy.short,
    "ushort": numpy.ushort,
    "double": numpy.double,
    "ulong": numpy.uint64,
    "ubyte": numpy.uint8,
    "byte": numpy.int8
}

_type_str = {
    "float": "f4",
    "int": "i4",
    "uint": "u4",
    "double": "f8",
    "byte": "i1",
    "ubyte": "u1",
    "short": "i2",
    "ushort": "u2",
    "string": "c",
    "char": "c",
    "long": "i4",
    "ulong": "u4",
    "String": "c",  # FIXME: DOE SEISS 0x420 has this?
}


def ncml_tag(t):
    return "{http://www.unidata.ucar.edu/namespaces/netcdf/ncml-2.2}%s" % t


def _process_dimension(xml_parser, nc, elem, event="start", parent=None):
    if elem.get("name") not in nc.dimensions:
        new_dim = nc.createDimension(elem.get("name"), elem.get(
            "length") and int(elem.get("length")))

    next_event, next_elem = xml_parser.next()
    while not (next_event == "end" and
               next_elem.tag == ncml_tag("dimension") and
               next_elem.get("name") == elem.get("name")):
        _process_element(xml_parser, nc, next_elem,
                         event=next_event, parent=new_dim)
        next_event, next_elem = xml_parser.next()

    return


def _process_values(xml_parser, nc, elem, event="start", parent=None):

    # for <values increment="..." start="..."/> tags
    if elem.get("start") is not None and elem.get("increment") is not None:
        start = int(elem.get("start"))
        stop = int(len(parent))
        increment = int(elem.get("increment"))
        parent.set_auto_maskandscale(False)
        # this will scale if we don't turn off scaling above
        parent[:] = range(start, stop, increment)

    if elem.text is None:
        return
    txt = elem.text.strip().split(" ")
    # we have to handle strings differently to split them character-wise
    # properly
    if parent.dtype == numpy.dtype('c'):
        val = numpy.chararray(parent.shape)
        # FIXME: this is really gross and can undoubtedly be improved, but
        # works for now...
        if len(parent.shape) == 1:
            # FIXME: why does looping like this work but not assigning the
            # whole thing at once?
            for i in range(len(txt[0])):
                val[i] = txt[0][i]
        elif len(parent.shape) == 2:
            try:
                for i in range(len(txt)):
                    toassign = list(txt[i])
                    toassign = toassign + [''] * (len(val[i]) - len(toassign))
                    val[i][:] = toassign
            except:
                import pdb; pdb.set_trace()
        else:
            raise RuntimeError, "something we've never seen before"
    else:
        val = numpy.array(txt, dtype=parent.dtype)
        if len(val) == 1:
            val = val[0]
    if parent.name in [u'lowWavelengthLines', u'highWavelengthLines']:
        parent[0,:] = val
    else:
        parent[:] = val
    return


def _process_attribute(xml_parser, nc, elem, event="start", parent=None):
        # Handle attribute type and value
    # _FillValue needs to be set at variable creation
    if not elem.get('name') == '_FillValue':
        val = elem.get("value")
        # this is a better fix than using splittable attrs I think
        if not elem.get("type") == 'string':
            val = val.split(" ")
        try:
            setattr(parent, elem.get("name"), _types[elem.get("type")](val))
        except:
            log.error("could not cast %s, data likely doesn't match the specified type" % (
                elem.get("name")))
            pass

    next_event, next_elem = xml_parser.next()
    while not (next_event == "end" and
               next_elem.tag == ncml_tag("attribute") and
               next_elem.get("name") == elem.get("name")):
            # FIXME: I don't know if parent=nc would be correct here
        _process_element(xml_parser, nc, next_elem,
                         event=next_event, parent=nc)
        next_event, next_elem = xml_parser.next()

    return


def _process_variable(xml_parser, nc, elem, event="start", parent=None):
    fill = None

    for attr in elem:
        if attr.get("name") == "_FillValue":
            fill = _types[attr.get("type")](attr.get("value"))

    if elem.get("shape"):
        dims = tuple(elem.get("shape").split(" "))

        while "" in dims:
            dims = list(dims)
            dims.remove("")
            dims = tuple(dims)

        # FIXME: NCML typo for MAG
        if "number_of_samples_per_report" in dims:
            dims = list(dims)
            dims[dims.index("number_of_samples_per_report")
                 ] = "number_samples_per_report"
            dims = tuple(dims)

        try:
            if elem.get("name") not in nc.variables:
                new_var = nc.createVariable(elem.get("name"), _type_str[
                                            elem.get("type")], dims, fill_value=fill,
                                            zlib=True, complevel=CSPP_GEO_GRBR_DEFAULT_COMPRESSION)
            else:
                new_var = nc.variables[elem.get("name")]

        except ValueError:
            log.error("could not add variable %s to the file, likely an invalid dimension in the ncml" % (
                elem.get("name")))
            return None
    else:
        if elem.get("name") not in nc.variables:
            new_var = nc.createVariable(elem.get("name"), _type_str[
                                        elem.get("type")], fill_value=fill,
                                        zlib=True, complevel=CSPP_GEO_GRBR_DEFAULT_COMPRESSION)
        else:
            new_var = nc.variables[elem.get("name")]

    # this handles values provided inside the <variable> tag
    if elem.get("values"):
        new_var[:] = elem.get("values")
# FIXME: this was zeroing out Rad/DQF when the metadata arrived, and since it arrives last, that zeros out all our data
    # otherwise we zero out and trust later lines to fill in the values as necessary (see _process_values)
#    else:
#        new_var.set_auto_maskandscale(False)
#        new_var[:] = 0

    next_event, next_elem = xml_parser.next()
    while not (next_event == "end" and
               next_elem.tag == ncml_tag("variable") and
               next_elem.get("name") == elem.get("name")):
        _process_element(xml_parser, nc, next_elem,
                         event=next_event, parent=new_var)
        next_event, next_elem = xml_parser.next()

    return None


def _process_element(xml_parser, nc, elem, event="start", parent=None):
    if elem.tag == "end":
        log.warning(
            "Something went wrong, we shouldn't be parsing an 'end' tag")
        return
    if elem.tag == ncml_tag("attribute"):
        _process_attribute(xml_parser, nc, elem, event=event, parent=parent)
    elif elem.tag == ncml_tag("variable"):
        _process_variable(xml_parser, nc, elem, event=event, parent=parent)
    elif elem.tag == ncml_tag("values"):
        _process_values(xml_parser, nc, elem, event=event, parent=parent)
    elif elem.tag == ncml_tag("netcdf"):
        pass
    elif elem.tag == ncml_tag("dimension"):
        pass  # because we handle these first, since they can be out of order in DOE data
    else:
        log.error("Unknown NCML element found (%r)" % (elem))
        raise ValueError("Unknown NCML element found (%r)" % (elem))


def create_nc_from_ncml(nc_filename, ncml_filename, format="NETCDF4"):
    """Take a NCML file and create a NetCDF file filled with the attributes
    and variables listed in the NCML file.
    """
    nc = Dataset(nc_filename, 'w', format=format)
    update_nc_from_ncml(nc, ncml_filename, format=format)


def update_nc_from_ncml(nc, ncml_filename, format="NETCDF4"):
    xml_parser = cElementTree.iterparse(ncml_filename, events=("start", "end"))
    for event, elem in xml_parser:  # FIXME: dimensions out of order in DOE data, handle these first
        if elem.tag == ncml_tag("dimension"):
            _process_dimension(xml_parser, nc, elem, event=event, parent=nc)

    xml_parser = cElementTree.iterparse(ncml_filename, events=("start", "end"))
    for event, elem in xml_parser:
        _process_element(xml_parser, nc, elem, event=event, parent=nc)

    return nc


def main():
    logging.basicConfig(level=logging.DEBUG)
    return create_nc_from_ncml(sys.argv[2], sys.argv[1])

if __name__ == "__main__":
    sys.exit(main())
